/*
	saori.cpp
    Multi Platform SAORI Framework
*/

#include "saori.h"

#ifdef _WINDOWS
#ifdef _DEBUG
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)
#endif /* _DEBUG */
#endif /* _WINDOWS */

// B SAORI CX^X
static boost::scoped_ptr<SAORI> saori;

bool SAORIInput::deserialize(string_t req)
{
	istringstream_t stream(req);
	string_t line;

	std::getline(stream, line);
	string_t::size_type proto_pos = line.find(_T(" SAORI/1."));
	if (proto_pos == string_t::npos)
	{
		return false;
	}

	command = line.substr(0, proto_pos);

	while (std::getline(stream, line))
	{
		string_t::size_type pos = line.find(_T(": "));
		if (pos != string_t::npos)
		{
			string_t key = line.substr(0, pos);
			string_t value = line.substr(pos + 2, line.size() - pos - 3);

			if (key.compare(0, 8, _T("Argument")) == 0)
			{

#ifdef _SAORI_UNICODE
				unsigned int order = wtoi(key.c_str() + 8);
#else
				unsigned int order = atoi(key.c_str() + 8);
#endif /* _SAORI_UNICODE */

                if (order == 0 && key[8] == _T('0'))
                {
                    //function = value;
                    if (args.size() < 1)
					{
						args.resize(1);
					}
					args[0] = value;
                }
				else if (order > 0)
				{
					if (args.size() < order+1)
					{
						args.resize(order+1);
					}
					args[order] = value;
				}
			}
			else
			{
				if (pos > 0)
				{
					opts[key] = value;
				}
			}
		}
	}
	return true;
}

string_t SAORIOutput::serialize()
{
	string_t rc_str = SAORI::ResultToString(result_code);

	ostringstream_t res;
	res << SAORI_VERSIONSTRING _T(" ");
	res << result_code << _T(" ") << rc_str << _T("\r\n");

	res << _T("Charset: ") << SAORI::CharsetToString(charset) << _T("\r\n");

    if (!result.empty())
    {
	    res << _T("Result: ") << result << _T("\r\n");
    }

	if (!values.empty())
	{
		for (int i = 0;i < (int)values.size(); ++i)
		{
			res << _T("Value") << i << _T(": ");

			string_t value = values[i];
			string_t::size_type pos = 0;
			while ((pos = value.find_first_of(_T("\r\n"), pos)) != string_t::npos)
			{
				value.replace(pos, 2, _T("\1"));
			}

			res << value << _T("\r\n");
		}
	}

	if (!opts.empty())
	{
		for (std::map<string_t, string_t>::const_iterator it = opts.begin();it != opts.end(); ++it)
		{
			res << it->first << _T(": ") << it->second << _T("\r\n");
		}
	}

	res << _T("\r\n");
	return res.str();
}

std::string SAORI::request(std::string req)
{
	string_t req_t;

	SAORICharset charset = SAORICHARSET_SHIFT_JIS;

#ifdef _SAORI_UNICODE
	std::string temp = req;
	std::transform(temp.begin(), temp.end(), temp.begin(), tolower);

	std::string::size_type char_pos = temp.find("\ncharset: ");
	if (char_pos != std::string::npos)
	{
		char_pos += 10;
		charset = SAORI::StringToCharset(temp.substr(char_pos, temp.find("\r\n", char_pos) - char_pos));
	}

	req_t = MultibyteToUnicode(charset, req);
#else
	req_t = req;
#endif /* _SAORI_UNICODE */

	SAORIInput in(req_t);
	in.charset = charset;
	in.opts[_T("SecurityLevel")] = _T("Local");

    SAORIOutput out;
	out.charset = in.charset;
	out.result_code = SAORIRESULT_INTERNAL_SERVER_ERROR;

	if (in.command == _T("GET Version"))
	{
		out.result_code = SAORIRESULT_OK;
	}
	else if (in.command == _T("EXECUTE"))
	{
        /*std::map<string_t, SAORIFunction>::const_iterator it = functions.lower_bound(in.function);
        if (it != functions.end())
        {
            out.result_code = it->second(in, out);
        }
        else
        {
            out.result_code = SAORIRESULT_BAD_REQUEST;
        }*/
        out.result_code = execute(in,out);
	}
	else
	{
		out.result_code = SAORIRESULT_BAD_REQUEST;
	}

	std::string res;
	string_t res_t = out.serialize();

#ifdef _SAORI_UNICODE
	res = UnicodeToMultibyte(out.charset, res_t);
#else
	res = res_t;
#endif /* _SAORI_UNICODE */

	return res;
}

string_t SAORI::ResultToString(SAORIResult result)
{
	switch (result)
	{
		case SAORIRESULT_OK:
			return _T("OK");
		case SAORIRESULT_NO_CONTENT:
			return _T("No Content");
		case SAORIRESULT_BAD_REQUEST:
			return _T("Bad Request");
		case SAORIRESULT_INTERNAL_SERVER_ERROR:
			return _T("Internal Server Error");
	}
	return _T("unknown error");
}

string_t SAORI::CharsetToString(SAORICharset charset)
{
	switch (charset)
	{
		case SAORICHARSET_SHIFT_JIS:
			return _T("Shift_JIS");
		case SAORICHARSET_ISO_2022_JP:
			return _T("ISO-2022-JP");
		case SAORICHARSET_EUC_JP:
			return _T("EUC-JP");
		case SAORICHARSET_UTF_8:
			return _T("UTF-8");
	}
	return _T("unknown charset");
}

SAORICharset SAORI::StringToCharset(const std::string &str)
{
	if (str == "shift_jis" || str =="x-sjis")
	{
		return SAORICHARSET_SHIFT_JIS;
	}
	else if (str == "iso-2022-jp")
	{
		return SAORICHARSET_ISO_2022_JP;
	}
	else if (str == "euc-jp" || str == "x-euc-jp")
	{
		return SAORICHARSET_EUC_JP;
	}
	else if (str == "utf-8")
	{
		return SAORICHARSET_UTF_8;
	}
	return SAORICHARSET_SHIFT_JIS;
}

#ifdef _SAORI_UNICODE

std::wstring SAORI::MultibyteToUnicode(SAORICharset charset, const std::string &str)
{
#ifdef _WINDOWS
	int length = MultiByteToWideChar(charset, 0, str.c_str(), -1, NULL, 0);
	std::vector<wchar_t> wstr(length + 1);
	length = MultiByteToWideChar(charset, 0, str.c_str(), -1, &wstr[0], length + 1);
	return std::wstring(wstr.begin(), wstr.begin() + length);
#else
	// not implemented
	return std::wstring();
#endif /* _WINDOWS */
}

std::string SAORI::UnicodeToMultibyte(SAORICharset charset, const std::wstring &wstr)
{
#ifdef _WINDOWS
	int length = WideCharToMultiByte(charset, 0, wstr.c_str(), -1, NULL, 0, NULL, NULL);
	std::vector<char> str(length + 1);
	length = WideCharToMultiByte(charset, 0, wstr.c_str(), -1, &str[0], length + 1, NULL, NULL);
	return std::string(str.begin(), str.begin() + length);
#else
	// not implemented
	return std::string();
#endif /* _WINDOWS */
}

#endif /* _SAORI_UNICODE */

#ifdef _WINDOWS

BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	// boost::scoped_ptr<T> gĂ̂ŉȂĂ悢
	return TRUE;
}

#endif /* _WINDOWS */

SAORIAPI int SAORICALL load(void *h, long len)
{
	if (saori)
	{
		unload();
	}
	if (!h)
	{
		return 0;
	}
	std::string path(reinterpret_cast<char *>(h), len);
	SAORI_FREE(h);
	saori.reset(new SAORI(path));
	return saori->load();
}

SAORIAPI int SAORICALL unload()
{
	if (!saori)
	{
		return 0;
	}
	int ret = saori->unload();
	saori.reset();
	return ret;
}

SAORIAPI void * SAORICALL request(void *h, long *len)
{
	std::string req(reinterpret_cast<char *>(h), *len);
	SAORI_FREE(h);

	std::string res = saori->request(req);

	*len = static_cast<long>(res.size());
	h = SAORI_ALLOC(*len + 1);
	if (h)
	{
		memcpy(h, res.c_str(), *len + 1);
	}
	return h;
}